#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script builds a COLMAP model (map) from kapture format (images, cameras, trajectories: no features or matches!)
"""

import argparse
import logging
import os
import os.path as path
import shutil
from typing import List

import path_to_kapture
import kapture
import kapture.io.csv
import kapture.utils.logging
from kapture.utils.paths import safe_remove_file, safe_remove_any_path
from kapture.core.Trajectories import rigs_remove_inplace
from kapture.io.records import get_image_fullpath

import kapture.converter.colmap.colmap_command as colmap_lib
from kapture.converter.colmap.database import COLMAPDatabase
import kapture.converter.colmap.database_extra as database_extra

logger = logging.getLogger('colmap_build_sift_map')


def colmap_build_sift_map(kapture_path: str,
                          colmap_path: str,
                          colmap_binary: str,
                          colmap_use_cpu: bool,
                          colmap_gpu_index: str,
                          vocab_tree_path: str,
                          point_triangulator_options: List[str],
                          skip_list: List[str],
                          force: bool) -> None:
    """
    Build a colmap model using default SIFT features with the kapture data.

    :param kapture_path: path to the kapture to use
    :param colmap_path: path to the colmap build
    :param colmap_binary: path to the colmap executable
    :param colmap_use_cpu: to use cpu only (and ignore gpu) or to use also gpu
    :param colmap_gpu_index: gpu index for sift extractor and mapper
    :param vocab_tree_path: path to the colmap vocabulary tree file
    :param point_triangulator_options: options for the point triangulator
    :param skip_list: list of steps to skip
    :param force: Silently overwrite kapture files if already exists.
    """
    os.makedirs(colmap_path, exist_ok=True)

    # Load input files first to make sure it is OK
    logger.info('loading kapture files...')
    kapture_data = kapture.io.csv.kapture_from_dir(kapture_path)

    if not (kapture_data.records_camera and kapture_data.sensors):
        raise ValueError('records_camera, sensors are mandatory')
    if not kapture_data.trajectories:
        logger.info('there are no trajectories, running mapper instead of point_triangulator')

    if not os.path.isfile(vocab_tree_path):
        raise ValueError(f'Vocabulary Tree file does not exist: {vocab_tree_path}')

    # COLMAP does not fully support rigs.
    if kapture_data.rigs is not None and kapture_data.trajectories is not None:
        # make sure, rigs are not used in trajectories.
        logger.info('remove rigs notation.')
        rigs_remove_inplace(kapture_data.trajectories, kapture_data.rigs)
        kapture_data.rigs.clear()

    # Set fixed name for COLMAP database
    colmap_db_path = path.join(colmap_path, 'colmap.db')
    image_list_path = path.join(colmap_path, 'images.list')
    reconstruction_path = path.join(colmap_path, "reconstruction")
    if 'delete_existing' not in skip_list:
        safe_remove_file(colmap_db_path, force)
        safe_remove_file(image_list_path, force)
        safe_remove_any_path(reconstruction_path, force)
    os.makedirs(reconstruction_path, exist_ok=True)

    if 'feature_extract' not in skip_list:
        logger.info("Step 1: Feature extraction using colmap")
        with open(image_list_path, 'w') as fid:
            for timestamp, sensor_id in sorted(kapture_data.records_camera.key_pairs()):
                fid.write(kapture_data.records_camera[timestamp][sensor_id] + "\n")

        colmap_lib.run_feature_extractor(
            colmap_binary,
            colmap_use_cpu,
            colmap_gpu_index,
            colmap_db_path,
            get_image_fullpath(kapture_path),
            image_list_path
        )

    # Update cameras in COLMAP:
    # - use only one camera for all images taken with the same camera (update all camera IDs)
    # - import camera intrinsics
    # - import camera pose
    if 'update_db_cameras' not in skip_list:
        logger.info("Step 2: Populate COLMAP DB with cameras and poses")
        colmap_db = COLMAPDatabase.connect(colmap_db_path)
        database_extra.update_DB_cameras_and_poses(colmap_db, kapture_data)
        # close db before running colmap processes in order to avoid locks
        colmap_db.close()

    # Extract matches with COLMAP
    if 'matches' not in skip_list:
        logger.info("Step 3: Compute matches with colmap")

        colmap_lib.run_vocab_tree_matcher(
            colmap_binary,
            colmap_use_cpu,
            colmap_gpu_index,
            colmap_db_path,
            vocab_tree_path)

    if kapture_data.trajectories is not None:
        # Generate priors for reconstruction
        txt_path = path.join(colmap_path, "priors_for_reconstruction")
        os.makedirs(txt_path, exist_ok=True)
        if 'priors_for_reconstruction' not in skip_list:
            logger.info('Step 4: Exporting priors for reconstruction.')
            colmap_db = COLMAPDatabase.connect(colmap_db_path)
            database_extra.generate_priors_for_reconstruction(kapture_data, colmap_db, txt_path)
            colmap_db.close()

        # Point triangulator
        reconstruction_path = path.join(colmap_path, "reconstruction")
        os.makedirs(reconstruction_path, exist_ok=True)
        if 'triangulation' not in skip_list:
            logger.info("Step 5: Triangulation")
            colmap_lib.run_point_triangulator(
                colmap_binary,
                colmap_db_path,
                get_image_fullpath(kapture_path),
                txt_path,
                reconstruction_path,
                point_triangulator_options
            )
    else:
        # mapper
        reconstruction_path = path.join(colmap_path, "reconstruction")
        os.makedirs(reconstruction_path, exist_ok=True)
        if 'triangulation' not in skip_list:
            logger.info("Step 5: Triangulation")
            colmap_lib.run_mapper(
                colmap_binary,
                colmap_db_path,
                get_image_fullpath(kapture_path),
                None,
                reconstruction_path,
                point_triangulator_options
            )
            # use reconstruction 0 as main
            first_reconstruction = os.path.join(reconstruction_path, '0')
            files = os.listdir(first_reconstruction)
            for f in files:
                shutil.move(os.path.join(first_reconstruction, f), os.path.join(reconstruction_path, f))
            shutil.rmtree(first_reconstruction)

    # run model_converter
    if 'model_converter' not in skip_list:
        logger.info("Step 6: Export reconstruction results to txt")
        colmap_lib.run_model_converter(
            colmap_binary,
            reconstruction_path,
            reconstruction_path
        )


def colmap_build_sift_map_command_line():
    """
    Parse the command line arguments to build a colmap map using the given kapture data and the sift feature.
    """
    parser = argparse.ArgumentParser(description='create a Colmap model (map) from data specified in kapture format.'
                                                 'Only images and cameras are taken into account '
                                                 ' (no features or matches)')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument('-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
                                  action=kapture.utils.logging.VerbosityParser,
                                  help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument('-q', '--silent', '--quiet', action='store_const',
                                  dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='silently delete database if already exists.')
    parser.add_argument('-i', '--input', required=True,
                        help='input path to kapture data root directory')
    parser.add_argument('-o', '--output', required=True,
                        help='output directory (colmap directory).')
    parser.add_argument('-colmap', '--colmap_binary', required=False,
                        default="colmap",
                        help='full path to colmap binary '
                             '(default is "colmap", i.e. assume the binary'
                             ' is in the user PATH).')
    parser.add_argument('--cpu', action='store_true',
                        default=False,
                        help='Ignore gpu and run on cpu only')
    parser.add_argument('-gpu_idx', '--colmap_gpu_index', required=False,
                        default=None,
                        help='GPU index used by colmap Sift extractor and matcher (e.g. "0" or "0,1,2").'
                             ' Default: None. Ignored if --cpu is used')
    parser.add_argument('-voc', '--vocab_tree_path', required=True,
                        help='full path to Vocabulary Tree file'
                             ' used for matching.')
    parser.add_argument('-s', '--skip', choices=['delete_existing',
                                                 'feature_extract',
                                                 'update_db_cameras',
                                                 'matches',
                                                 'priors_for_reconstruction',
                                                 'triangulation',
                                                 'model_converter'],
                        nargs='+', default=[],
                        help='steps to skip')

    args, point_triangulator_options = parser.parse_known_args()

    logger.setLevel(args.verbose)
    logging.getLogger('colmap').setLevel(args.verbose)
    if args.verbose <= logging.INFO:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    args_dict = vars(args)
    args_dict['point_triangulator_options'] = point_triangulator_options
    logger.debug('colmap_build_sift_map.py \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v)) for k, v in args_dict.items()))

    colmap_build_sift_map(args.input, args.output, args.colmap_binary,
                          args.cpu, args.colmap_gpu_index,
                          args.vocab_tree_path, point_triangulator_options,
                          args.skip, args.force)


if __name__ == '__main__':
    colmap_build_sift_map_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to import a 7scenes model into a kapture.
The RGB-D Dataset 7-Scenes data structure is defined here:
    https://www.microsoft.com/en-us/research/project/rgb-d-dataset-7-scenes/

    Each sequence (seq-XX.zip) consists of 500-1000 frames. Each frame consists of three files:
     - Color: frame-XXXXXX.color.png (RGB, 24-bit, PNG)
     - Depth: frame-XXXXXX.depth.png (depth in millimeters, 16-bit, PNG, invalid depth is set to 65535).
     - Pose: frame-XXXXXX.pose.txt (camera-to-world, 4×4 matrix in homogeneous coordinates).

"""

import argparse
import logging
import os
import os.path as path
import re
import numpy as np
import quaternion
from PIL import Image
# kapture
import path_to_kapture
import kapture
import kapture.utils.logging
from kapture.io.structure import delete_existing_kapture_files
from kapture.io.csv import kapture_to_dir
import kapture.io.features
from kapture.io.records import TransferAction, import_record_data_from_dir_auto

logger = logging.getLogger('7scenes')
MODEL = kapture.CameraType.RADIAL


def import_7scenes(d7scenes_path: str,
                   kapture_dir_path: str,
                   force_overwrite_existing: bool = False,
                   images_import_method: TransferAction = TransferAction.skip) -> None:
    """
    Imports RGB-D Dataset 7-Scenes dataset and save them as kapture.

    :param d7scenes_path: path to the 7scenes sequence root path
    :param kapture_dir_path: path to kapture top directory
    :param force_overwrite_existing: Silently overwrite kapture files if already exists.
    :param images_import_method: choose how to import actual image files.
    """
    os.makedirs(kapture_dir_path, exist_ok=True)
    delete_existing_kapture_files(kapture_dir_path, force_erase=force_overwrite_existing)

    logger.info('loading all content ...')
    POSE_SUFFIX = 'pose'
    RGB_SUFFIX = 'color'
    DEPTH_SUFFIX = 'depth'
    CAMERA_ID = 'kinect'
    d7s_filename_re = re.compile(r'frame-(?P<timestamp>\d{6})\.(?P<suffix>\w*)\.(?P<ext>\w*)')

    # populate
    d7s_filenames = (path.basename(path.join(dp, fn))
                     for dp, _, fs in os.walk(d7scenes_path) for fn in fs)
    d7s_filenames = {filename: d7s_filename_re.match(filename).groupdict()
                     for filename in d7s_filenames
                     if d7s_filename_re.match(filename)}
    # d7s_filenames -> timestamp, suffix, ext
    if not d7s_filenames:
        raise ValueError('no pose file found: make sure the path to 7scenes sequence is valid.')

    # images
    logger.info('populating image files ...')
    d7s_filenames_images = ((int(v['timestamp']), filename)
                            for filename, v in d7s_filenames.items()
                            if v['suffix'] == RGB_SUFFIX)
    snapshots = kapture.RecordsCamera()
    for timestamp, image_filename in sorted(d7s_filenames_images):
        snapshots[timestamp, CAMERA_ID] = image_filename

    # poses
    logger.info('import poses files ...')
    d7s_filenames_poses = ((int(v['timestamp']), filename)
                           for filename, v in d7s_filenames.items()
                           if v['suffix'] == POSE_SUFFIX)
    trajectories = kapture.Trajectories()
    for timestamp, pose_filename in d7s_filenames_poses:
        pose_filepath = path.join(d7scenes_path, pose_filename)
        pose_mat = np.loadtxt(pose_filepath)  # camera-to-world, 4×4 matrix in homogeneous coordinates
        rotation_mat = pose_mat[0:3, 0:3]
        position_vec = pose_mat[0:3, 3]
        rotation_quat = quaternion.from_rotation_matrix(rotation_mat)
        pose_world_from_cam = kapture.PoseTransform(r=rotation_quat, t=position_vec)
        pose_cam_from_world = pose_world_from_cam.inverse()
        trajectories[timestamp, CAMERA_ID] = pose_cam_from_world

    # sensors
    """
    From authors: The RGB and depth camera have not been calibrated and we can’t provide calibration parameters at the 
    moment. The recorded frames correspond to the raw, uncalibrated camera images. In the KinectFusion pipeline we used 
    the following default intrinsics for the depth camera: Principle point (320,240), Focal length (585,585).
    """
    sensors = kapture.Sensors()
    sensors[CAMERA_ID] = kapture.Camera(
        name='kinect',
        camera_type=kapture.CameraType.SIMPLE_PINHOLE,
        camera_params=[640, 480, 585, 320, 240]  # w, h, f, cx, cy
    )

    # import (copy) image files.
    logger.info('copying image files ...')
    image_filenames = [f for _, _, f in kapture.flatten(snapshots)]
    import_record_data_from_dir_auto(d7scenes_path, kapture_dir_path, image_filenames, images_import_method)

    # pack into kapture format
    imported_kapture = kapture.Kapture(
        records_camera=snapshots,
        trajectories=trajectories,
        sensors=sensors)

    logger.info('writing imported data ...')
    kapture_to_dir(kapture_dir_path, imported_kapture)


def import_7scenes_command_line() -> None:
    """
    Imports RGB-D Dataset 7-Scenes dataset and save them as kapture using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(
        description='Imports RGB-D Dataset 7-Scenes files to the kapture format.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output if already exists.')
    # import ###########################################################################################################
    parser.add_argument('-i', '--input', required=True,
                        help='input path Dataset 7-Scenes sequence root path')
    parser.add_argument('--image_transfer', type=TransferAction, default=TransferAction.link_absolute,
                        help=f'How to import images [link_absolute], '
                             f'choose among: {", ".join(a.name for a in TransferAction)}')
    parser.add_argument('-o', '--output', required=True, help='output directory.')
    ####################################################################################################################
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    import_7scenes(d7scenes_path=args.input,
                   kapture_dir_path=args.output,
                   force_overwrite_existing=args.force,
                   images_import_method=args.image_transfer)


if __name__ == '__main__':
    import_7scenes_command_line()



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script builds a COLMAP model (map) from kapture format (images, cameras, trajectories, features and matches!)
"""

import argparse
import logging
import os
import os.path as path
import shutil
from typing import List, Optional

import path_to_kapture
import kapture
import kapture.io.csv
import kapture.utils.logging
from kapture.utils.paths import safe_remove_file, safe_remove_any_path
from kapture.core.Trajectories import rigs_remove_inplace
from kapture.io.records import get_image_fullpath

import kapture.converter.colmap.colmap_command as colmap_lib
from kapture.converter.colmap.database import COLMAPDatabase
import kapture.converter.colmap.database_extra as database_extra

logger = logging.getLogger('colmap_build_map')


def colmap_build_map(kapture_path: str,
                     colmap_path: str,
                     colmap_binary: str,
                     pairsfile_path: Optional[str],
                     use_colmap_matches_importer: bool,
                     point_triangulator_options: List[str],
                     skip_list: List[str],
                     force: bool) -> None:
    """
    Build a colmap model using custom features with the kapture data.

      :param kapture_path: path to the kapture to use
      :param colmap_path: path to the colmap build
      :param colmap_binary: path to the colmap executable
      :param pairsfile_path: Optional[str],
      :param use_colmap_matches_importer: bool,
      :param point_triangulator_options: options for the point triangulator
      :param skip_list: list of steps to skip
      :param force: Silently overwrite kapture files if already exists.
      """
    # Load input files first to make sure it is OK
    logger.info('loading kapture files...')
    kapture_data = kapture.io.csv.kapture_from_dir(kapture_path, matches_pairsfile_path=pairsfile_path)

    colmap_build_map_from_loaded_data(kapture_data,
                                      kapture_path,
                                      colmap_path,
                                      colmap_binary,
                                      pairsfile_path,
                                      use_colmap_matches_importer,
                                      point_triangulator_options,
                                      skip_list,
                                      force)


def colmap_build_map_from_loaded_data(kapture_data: kapture.Kapture,
                                      kapture_path: str,
                                      colmap_path: str,
                                      colmap_binary: str,
                                      pairsfile_path: Optional[str],
                                      use_colmap_matches_importer: bool,
                                      point_triangulator_options: List[str],
                                      skip_list: List[str],
                                      force: bool) -> None:
    """
    Build a colmap model using custom features with the kapture data.

    :param kapture_data: kapture data to use
    :param kapture_path: path to the kapture to use
    :param colmap_path: path to the colmap build
    :param colmap_binary: path to the colmap executable
    :param pairsfile_path: Optional[str],
    :param use_colmap_matches_importer: bool,
    :param point_triangulator_options: options for the point triangulator
    :param skip_list: list of steps to skip
    :param force: Silently overwrite kapture files if already exists.
    """
    os.makedirs(colmap_path, exist_ok=True)

    if not (kapture_data.records_camera and kapture_data.sensors and kapture_data.keypoints and kapture_data.matches):
        raise ValueError('records_camera, sensors, keypoints, matches are mandatory')
    if not kapture_data.trajectories:
        logger.info('there are no trajectories, running mapper instead of point_triangulator')

    # COLMAP does not fully support rigs.
    if kapture_data.rigs is not None and kapture_data.trajectories is not None:
        # make sure, rigs are not used in trajectories.
        logger.info('remove rigs notation.')
        rigs_remove_inplace(kapture_data.trajectories, kapture_data.rigs)
        kapture_data.rigs.clear()

    # Set fixed name for COLMAP database
    colmap_db_path = path.join(colmap_path, 'colmap.db')
    reconstruction_path = path.join(colmap_path, "reconstruction")
    priors_txt_path = path.join(colmap_path, "priors_for_reconstruction")
    if 'delete_existing' not in skip_list:
        safe_remove_file(colmap_db_path, force)
        safe_remove_any_path(reconstruction_path, force)
        safe_remove_any_path(priors_txt_path, force)
    os.makedirs(reconstruction_path, exist_ok=True)

    if 'colmap_db' not in skip_list:
        logger.info('Using precomputed keypoints and matches')
        logger.info('Step 1: Export kapture format to colmap')

        colmap_db = COLMAPDatabase.connect(colmap_db_path)
        if kapture_data.descriptors is not None:
            kapture_data.descriptors.clear()
        database_extra.kapture_to_colmap(kapture_data, kapture_path, colmap_db,
                                         export_two_view_geometry=not use_colmap_matches_importer)
        # close db before running colmap processes in order to avoid locks
        colmap_db.close()

        if use_colmap_matches_importer:
            logger.info('Step 2: Run geometric verification')
            logger.debug('running colmap matches_importer...')
            colmap_lib.run_matches_importer_from_kapture(
                colmap_binary,
                colmap_use_cpu=True,
                colmap_gpu_index=None,
                colmap_db_path=colmap_db_path,
                kapture_data=kapture_data,
                force=force
            )
        else:
            logger.info('Step 2: Run geometric verification - skipped')

    if kapture_data.trajectories is not None:
        # Generate priors for reconstruction
        os.makedirs(priors_txt_path, exist_ok=True)
        if 'priors_for_reconstruction' not in skip_list:
            logger.info('Step 3: Exporting priors for reconstruction.')
            colmap_db = COLMAPDatabase.connect(colmap_db_path)
            database_extra.generate_priors_for_reconstruction(kapture_data, colmap_db, priors_txt_path)
            colmap_db.close()

        # Point triangulator
        reconstruction_path = path.join(colmap_path, "reconstruction")
        os.makedirs(reconstruction_path, exist_ok=True)
        if 'triangulation' not in skip_list:
            logger.info("Step 4: Triangulation")
            colmap_lib.run_point_triangulator(
                colmap_binary,
                colmap_db_path,
                get_image_fullpath(kapture_path),
                priors_txt_path,
                reconstruction_path,
                point_triangulator_options
            )
    else:
        # mapper
        reconstruction_path = path.join(colmap_path, "reconstruction")
        os.makedirs(reconstruction_path, exist_ok=True)
        if 'triangulation' not in skip_list:
            logger.info("Step 4: Triangulation")
            colmap_lib.run_mapper(
                colmap_binary,
                colmap_db_path,
                get_image_fullpath(kapture_path),
                None,
                reconstruction_path,
                point_triangulator_options
            )
            # use reconstruction 0 as main
            first_reconstruction = os.path.join(reconstruction_path, '0')
            files = os.listdir(first_reconstruction)
            for f in files:
                shutil.move(os.path.join(first_reconstruction, f), os.path.join(reconstruction_path, f))
            shutil.rmtree(first_reconstruction)

    # run model_converter
    if 'model_converter' not in skip_list:
        logger.info("Step 5: Export reconstruction results to txt")
        colmap_lib.run_model_converter(
            colmap_binary,
            reconstruction_path,
            reconstruction_path
        )


def colmap_build_map_command_line():
    """
    Parse the command line arguments to build a colmap map using the given kapture data.
    """
    parser = argparse.ArgumentParser(description=('create a Colmap model (map) from data specified in kapture format.'
                                                  'kapture data must contain keypoints and matches'))
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument('-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
                                  action=kapture.utils.logging.VerbosityParser,
                                  help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument('-q', '--silent', '--quiet', action='store_const',
                                  dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='silently delete database if already exists.')
    parser.add_argument('-i', '--input', required=True,
                        help='input path to kapture data root directory')
    parser.add_argument('-o', '--output', required=True,
                        help='output directory (colmap directory).')
    parser.add_argument('-colmap', '--colmap_binary', required=False,
                        default="colmap",
                        help='full path to colmap binary '
                             '(default is "colmap", i.e. assume the binary'
                             ' is in the user PATH).')
    parser.add_argument('--use-colmap-matches-importer', action='store_true', default=False,
                        help='Use colmap matches_importer instead of manually filling the two_view_geometry table')
    parser.add_argument('--pairsfile-path',
                        default=None,
                        type=str,
                        help=('text file in the csv format; where each line is image_name1, image_name2, score '
                              'which contains the image pairs to match, can be used to filter loaded matches'))
    parser.add_argument('-s', '--skip', choices=['delete_existing',
                                                 'colmap_db',
                                                 'priors_for_reconstruction',
                                                 'triangulation',
                                                 'model_converter'],
                        nargs='+', default=[],
                        help='steps to skip')

    args, point_triangulator_options = parser.parse_known_args()

    logger.setLevel(args.verbose)
    logging.getLogger('colmap').setLevel(args.verbose)
    if args.verbose <= logging.INFO:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    args_dict = vars(args)
    args_dict['point_triangulator_options'] = point_triangulator_options
    logger.debug('colmap_build_map.py \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v)) for k, v in args_dict.items()))

    colmap_build_map(args.input, args.output, args.colmap_binary,
                     args.pairsfile_path, args.use_colmap_matches_importer, point_triangulator_options,
                     args.skip, args.force)


if __name__ == '__main__':
    colmap_build_map_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
All kapture objects representing the data kapture manages.
"""

from .PoseTransform import PoseTransform
from .Sensors import Sensor, Sensors, create_sensor
from .Sensors import Camera, CameraType, CAMERA_TYPE_PARAMS_COUNT, CAMERA_TYPE_PARAMS_COUNT_FROM_NAME
from .Rigs import Rigs
from .Trajectories import Trajectories, rigs_remove, rigs_remove_inplace, rigs_recover, rigs_recover_inplace
from .Records import RecordsBase, RecordsCamera, RecordsLidar, RecordsWifi, RecordWifi, RecordGnss, RecordsGnss
from .ImageFeatures import Keypoints, Descriptors, GlobalFeatures
from .Observations import Observations
from .Matches import Matches
from .Points3d import Points3d
from .Kapture import Kapture
from .flatten import flatten

import logging

# it is up to the library client to specify the log handler.
logging.getLogger("kapture").addHandler(logging.NullHandler())



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Script to export a kapture into an openmvg file.
"""

import argparse
import logging
import sys
# kapture
import path_to_kapture  # enables import kapture
import kapture.utils.logging
from kapture.io.records import TransferAction
from kapture.converter.openmvg.export_openmvg import export_openmvg

logger = logging.getLogger('openmvg')


def export_openmvg_command_line():
    """
    Do the kapture to openmvg export using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='Exports from kapture format to openMVG JSON file.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output file if already exists.')
    # create the parser for the export command #########################################################################
    parser.add_argument('-k', '--kapture', required=True,
                        help='input path to kapture data root directory')
    parser.add_argument('-o', '--openmvg',
                        help='output openMVG JSON file path.')
    parser.add_argument('--image_action', default='root_link', type=TransferAction,
                        help=f'''what to do with images:
        {TransferAction.root_link.name}: link to the root of the images directory (default) ;
        {TransferAction.link_absolute.name}: absolute individual file link ;
        {TransferAction.link_relative.name}: relative individual file link ;
        {TransferAction.copy.name}: copy file instead of creating link ;
        {TransferAction.move.name}: move file instead of creating link ;
        {TransferAction.skip.name}: do not create links
                                      ''')

    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # for debug, let kapture express itself.
        kapture.utils.logging.getLogger().setLevel(args.verbose)
    logger.debug(f'{sys.argv[0]} \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v))
        for k, v in vars(args).items()))

    # Check that we will not try to do symbolics links on Windows
    # because this is not possible today if the user is not admin
    if sys.platform.startswith("win") and \
            (args.image_link == TransferAction.link_relative or
             args.image_link == TransferAction.link_absolute or
             args.image_link == TransferAction.root_link):
        logger.fatal("It is currently impossible to link files on a Windows platform")
        logger.fatal(f"Please try another option for the image: either do nothing ({TransferAction.skip}),"
                     f" or {TransferAction.copy}")
        raise OSError("Image file linking not possible on Windows")
    else:
        export_openmvg(args.kapture, args.openmvg, args.image_action, args.force)


if __name__ == '__main__':
    try:
        export_openmvg_command_line()
    except Exception as any_ex:
        logger.error('Fatal error', exc_info=True)



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
This script localize images on an existing COLMAP model (map) from kapture format
images, cameras, trajectories: no features or matches!
"""

import argparse
import logging
import os
import os.path as path
import shutil
from typing import List, Optional

import path_to_kapture
import kapture
import kapture.io.csv
import kapture.utils.logging
from kapture.utils.paths import safe_remove_file, safe_remove_any_path
from kapture.core.Trajectories import rigs_remove_inplace

import kapture.converter.colmap.colmap_command as colmap_lib
from kapture.converter.colmap.database import COLMAPDatabase
import kapture.converter.colmap.database_extra as database_extra

logger = logging.getLogger('colmap_localize')


def colmap_localize(kapture_path: str,
                    colmap_path: str,
                    input_database_path: str,
                    input_reconstruction_path: str,
                    colmap_binary: str,
                    pairsfile_path: Optional[str],
                    use_colmap_matches_importer: bool,
                    image_registrator_options: List[str],
                    skip_list: List[str],
                    force: bool) -> None:
    """
    Localize images on a colmap model using default SIFT features with the kapture data.

    :param kapture_path: path to the kapture to use
    :param colmap_path: path to the colmap build
    :param input_database_path: path to the map colmap.db
    :param input_database_path: path to the map colmap.db
    :param input_reconstruction_path: path to the map reconstruction folder
    :param pairsfile_path: Optional[str],
    :param use_colmap_matches_importer: bool,
    :param image_registrator_options: options for the image registrator
    :param skip_list: list of steps to skip
    :param force: Silently overwrite kapture files if already exists.
    """
    # Load input files first to make sure it is OK
    logger.info('loading kapture files...')
    kapture_data = kapture.io.csv.kapture_from_dir(kapture_path, matches_pairsfile_path=pairsfile_path)

    colmap_localize_from_loaded_data(kapture_data,
                                     kapture_path,
                                     colmap_path,
                                     input_database_path,
                                     input_reconstruction_path,
                                     colmap_binary,
                                     pairsfile_path,
                                     use_colmap_matches_importer,
                                     image_registrator_options,
                                     skip_list,
                                     force)


def colmap_localize_from_loaded_data(kapture_data: kapture.Kapture,
                                     kapture_path: str,
                                     colmap_path: str,
                                     input_database_path: str,
                                     input_reconstruction_path: str,
                                     colmap_binary: str,
                                     pairsfile_path: Optional[str],
                                     use_colmap_matches_importer: bool,
                                     image_registrator_options: List[str],
                                     skip_list: List[str],
                                     force: bool) -> None:
    """
    Localize images on a colmap model using default SIFT features with the kapture data.

    :param kapture_data: kapture data to use
    :param kapture_path: path to the kapture to use
    :param colmap_path: path to the colmap build
    :param input_database_path: path to the map colmap.db
    :param input_database_path: path to the map colmap.db
    :param input_reconstruction_path: path to the map reconstruction folder
    :param pairsfile_path: Optional[str],
    :param use_colmap_matches_importer: bool,
    :param image_registrator_options: options for the image registrator
    :param skip_list: list of steps to skip
    :param force: Silently overwrite kapture files if already exists.
    """
    os.makedirs(colmap_path, exist_ok=True)

    if not (kapture_data.records_camera and kapture_data.sensors and kapture_data.keypoints and kapture_data.matches):
        raise ValueError('records_camera, sensors, keypoints, matches are mandatory')

    if kapture_data.trajectories:
        logger.warning("Input data contains trajectories: they will be ignored")
        kapture_data.trajectories.clear()
    else:
        kapture_data.trajectories = kapture.Trajectories()

    # COLMAP does not fully support rigs.
    if kapture_data.rigs is not None and kapture_data.trajectories is not None:
        # make sure, rigs are not used in trajectories.
        logger.info('remove rigs notation.')
        rigs_remove_inplace(kapture_data.trajectories, kapture_data.rigs)
        kapture_data.rigs.clear()

    # Prepare output
    # Set fixed name for COLMAP database
    colmap_db_path = path.join(colmap_path, 'colmap.db')
    image_list_path = path.join(colmap_path, 'images.list')
    reconstruction_path = path.join(colmap_path, "reconstruction")
    if 'delete_existing' not in skip_list:
        safe_remove_file(colmap_db_path, force)
        safe_remove_file(image_list_path, force)
        safe_remove_any_path(reconstruction_path, force)
    os.makedirs(reconstruction_path, exist_ok=True)

    # Copy colmap db to output
    if not os.path.exists(colmap_db_path):
        shutil.copy(input_database_path, colmap_db_path)

    # find correspondences between the colmap db and the kapture data
    images_all = {image_path: (ts, cam_id)
                  for ts, shot in kapture_data.records_camera.items()
                  for cam_id, image_path in shot.items()}

    colmap_db = COLMAPDatabase.connect(colmap_db_path)
    colmap_image_ids = database_extra.get_colmap_image_ids_from_db(colmap_db)
    colmap_images = database_extra.get_images_from_database(colmap_db)
    colmap_db.close()

    # dict ( kapture_camera -> colmap_camera_id )
    colmap_camera_ids = {images_all[image_path][1]: colmap_cam_id
                         for image_path, colmap_cam_id in colmap_images if image_path in images_all}

    images_to_add = {image_path: value
                     for image_path, value in images_all.items()
                     if image_path not in colmap_image_ids}

    flatten_images_to_add = [(ts, kapture_cam_id, image_path)
                             for image_path, (ts, kapture_cam_id) in images_to_add.items()]

    if 'import_to_db' not in skip_list:
        logger.info("Step 1: Add precomputed keypoints and matches to colmap db")
        cameras_to_add = kapture.Sensors()
        for _, (_, kapture_cam_id) in images_to_add.items():
            if kapture_cam_id not in colmap_camera_ids:
                kapture_cam = kapture_data.sensors[kapture_cam_id]
                cameras_to_add[kapture_cam_id] = kapture_cam
        colmap_db = COLMAPDatabase.connect(colmap_db_path)
        colmap_added_camera_ids = database_extra.add_cameras_to_database(cameras_to_add, colmap_db)
        colmap_camera_ids.update(colmap_added_camera_ids)

        colmap_added_image_ids = database_extra.add_images_to_database_from_flatten(
            colmap_db, flatten_images_to_add, kapture_data.trajectories, colmap_camera_ids)
        colmap_image_ids.update(colmap_added_image_ids)

        colmap_image_ids_reversed = {v: k for k, v in colmap_image_ids.items()}  # colmap_id : name

        # add new features
        colmap_keypoints = database_extra.get_keypoints_set_from_database(colmap_db, colmap_image_ids_reversed)

        keypoints_all = kapture_data.keypoints
        keypoints_to_add = {name for name in keypoints_all if name not in colmap_keypoints}
        keypoints_to_add = kapture.Keypoints(keypoints_all.type_name, keypoints_all.dtype, keypoints_all.dsize,
                                             keypoints_to_add)
        database_extra.add_keypoints_to_database(colmap_db, keypoints_to_add, kapture_path, colmap_image_ids)

        # add new matches
        colmap_matches = kapture.Matches(database_extra.get_matches_set_from_database(colmap_db,
                                                                                      colmap_image_ids_reversed))
        colmap_matches.normalize()

        matches_all = kapture_data.matches
        matches_to_add = kapture.Matches({pair for pair in matches_all if pair not in colmap_matches})
        # print(list(matches_to_add))
        database_extra.add_matches_to_database(colmap_db, matches_to_add, kapture_path, colmap_image_ids,
                                               export_two_view_geometry=not use_colmap_matches_importer)
        colmap_db.close()

    if use_colmap_matches_importer:
        logger.info('Step 2: Run geometric verification')
        logger.debug('running colmap matches_importer...')
        # compute two view geometry
        colmap_lib.run_matches_importer_from_kapture(
            colmap_binary,
            colmap_use_cpu=True,
            colmap_gpu_index=None,
            colmap_db_path=colmap_db_path,
            kapture_data=kapture_data,
            force=force)
    else:
        logger.info('Step 2: Run geometric verification - skipped')
    if 'image_registrator' not in skip_list:
        logger.info("Step 3: Run image_registrator")
        # run image_registrator
        colmap_lib.run_image_registrator(
            colmap_binary,
            colmap_db_path,
            input_reconstruction_path,
            reconstruction_path,
            image_registrator_options
        )

    # run model_converter
    if 'model_converter' not in skip_list:
        logger.info("Step 4: Export reconstruction results to txt")
        colmap_lib.run_model_converter(
            colmap_binary,
            reconstruction_path,
            reconstruction_path
        )


def colmap_localize_command_line():
    """
    Parse the command line arguments to localize images on an existing colmap map using the given kapture data.
    """
    parser = argparse.ArgumentParser(description=('localize images on a colmap model (map) '
                                                  'from data specified in kapture format.'
                                                  'Only images and cameras are taken into account '
                                                  ' (no features or matches)'))
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument('-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
                                  action=kapture.utils.logging.VerbosityParser,
                                  help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument('-q', '--silent', '--quiet', action='store_const',
                                  dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='silently delete database if already exists.')
    parser.add_argument('-i', '--input', required=True,
                        help='input path to kapture data root directory')
    parser.add_argument('-db', '--database',
                        help='path to COLMAP database file.')
    parser.add_argument('-txt', '--reconstruction',
                        help='path to COLMAP reconstruction triplet text file.')
    parser.add_argument('-o', '--output', required=True,
                        help='output directory (colmap directory).')
    parser.add_argument('-colmap', '--colmap_binary', required=False,
                        default="colmap",
                        help='full path to colmap binary '
                             '(default is "colmap", i.e. assume the binary'
                             ' is in the user PATH).')
    parser.add_argument('--use-colmap-matches-importer', action='store_true', default=False,
                        help='Use colmap matches_importer instead of manually filling the two_view_geometry table')
    parser.add_argument('--pairsfile-path',
                        default=None,
                        type=str,
                        help=('text file in the csv format; where each line is image_name1, image_name2, score '
                              'which contains the image pairs to match, can be used to filter loaded matches'))
    parser.add_argument('-s', '--skip', choices=['delete_existing',
                                                 'import_to_db',
                                                 'image_registrator',
                                                 'model_converter'],
                        nargs='+', default=[],
                        help='steps to skip')

    args, image_registrator_options = parser.parse_known_args()

    logger.setLevel(args.verbose)
    logging.getLogger('colmap').setLevel(args.verbose)
    if args.verbose <= logging.INFO:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)

    args_dict = vars(args)
    args_dict['image_registrator_options'] = image_registrator_options
    logger.debug('colmap_localize.py \\\n' + '  \\\n'.join(
        '--{:20} {:100}'.format(k, str(v)) for k, v in args_dict.items()))

    colmap_localize(args.input, args.output,
                    args.database, args.reconstruction,
                    args.colmap_binary,
                    args.pairsfile_path, args.use_colmap_matches_importer,
                    image_registrator_options,
                    args.skip, args.force)


if __name__ == '__main__':
    colmap_localize_command_line()



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Mainly about rig calibration
"""

import numpy as np
import quaternion
import math
from tqdm import tqdm
from typing import Optional, Tuple, Union, List
import logging
from kapture import Trajectories, Rigs
from kapture.utils.logging import getLogger

logger = getLogger()


def rigs_calibrate_average(
        trajectories: Trajectories,
        rigs: Rigs,
        master_sensors: Optional[List[str]] = None
) -> Rigs:
    """
    Recomputes rig calibrations, ie. the poses of the senors inside a rig.
    It uses the poses of the senors in the given trajectories, to average a average rig.

    :param trajectories:
    :param rigs:
    :param master_sensors:
    :return:
    """
    # senor_id -> rig_id
    sensors_to_rigs = {sensor_id: rig_id
                       for rig_id, sensor_id in rigs.key_pairs()}

    show_progress_bar = logger.getEffectiveLevel() <= logging.INFO
    for timestamp, poses in tqdm(trajectories.items(), disable=not show_progress_bar):
        # accumulate relative poses inside rig
        for sensor_id, pose_sensor_from_world in poses.items():
            pass

    rigs_calibrated = Rigs()
    return rigs_calibrated



# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Merge kapture objects.
"""

from typing import List, Optional, Type

import kapture
from kapture.io.records import TransferAction, get_image_fullpath

from .merge_reconstruction import merge_keypoints, merge_descriptors, merge_global_features, merge_matches
from .merge_reconstruction import merge_points3d_and_observations, merge_points3d
from .merge_records_data import merge_records_data


def merge_sensors(sensors_list: List[Optional[kapture.Sensors]]) -> kapture.Sensors:
    """
    Merge several sensors lists. For sensor with the same identifier, keep only the first one.

    :param sensors_list: list of sensors
    :return: merge sensors
    """
    assert len(sensors_list) > 0

    merged_sensors = kapture.Sensors()
    for sensors in sensors_list:
        if sensors is None:
            continue
        for sensor_id in sensors.keys():
            if sensor_id in merged_sensors:
                continue
            merged_sensors[sensor_id] = sensors[sensor_id]
    return merged_sensors


def merge_rigs(rigs_list: List[Optional[kapture.Rigs]]) -> kapture.Rigs:
    """
    Merge several rigs lists. For sensor with the same rig and sensor identifier, keep only the first one.

    :param rigs_list: list of rigs
    :return: merged rigs
    """
    assert len(rigs_list) > 0

    merged_rigs = kapture.Rigs()
    for rigs in rigs_list:
        if rigs is None:
            continue
        for rig_id, sensor_id in rigs.key_pairs():
            if rig_id in merged_rigs and sensor_id in merged_rigs[rig_id]:
                continue
            merged_rigs[(rig_id, sensor_id)] = rigs[(rig_id, sensor_id)]
    return merged_rigs


def merge_trajectories(trajectories_list: List[Optional[kapture.Trajectories]]) -> kapture.Trajectories:
    """
    Merge several trajectories lists. For trajectory point with the same timestamp and sensor identifier,
     keep only the first one.

    :param trajectories_list: list of trajectories
    :return: merged trajectories
    """
    assert len(trajectories_list) > 0

    merged_trajectories = kapture.Trajectories()
    for trajectories in trajectories_list:
        if trajectories is None:
            continue
        for timestamp, sensor_id, pose in kapture.flatten(trajectories):
            if (timestamp, sensor_id) in merged_trajectories:
                continue
            merged_trajectories[(timestamp, sensor_id)] = pose
    return merged_trajectories


def merge_records_camera(records_camera_list: List[Optional[kapture.RecordsCamera]]) -> kapture.RecordsCamera:
    """
    Merge several camera records lists. For camera record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_camera_list: list of camera records
    :return: merged camera records
    """
    assert len(records_camera_list) > 0

    merged_records_camera = kapture.RecordsCamera()
    for records_camera in records_camera_list:
        if records_camera is None:
            continue
        for timestamp, sensor_id, filename in kapture.flatten(records_camera):
            if (timestamp, sensor_id) in merged_records_camera:
                continue
            merged_records_camera[(timestamp, sensor_id)] = filename
    return merged_records_camera


def merge_records_lidar(records_lidar_list: List[Optional[kapture.RecordsLidar]]) -> kapture.RecordsLidar:
    """
    Merge several lidar records lists. For lidar record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_lidar_list: list of lidar records
    :return: merged lidar records
    """
    assert len(records_lidar_list) > 0

    merged_records_lidar = kapture.RecordsLidar()
    for records_lidar in records_lidar_list:
        if records_lidar is None:
            continue
        for timestamp, sensor_id, filename in kapture.flatten(records_lidar):
            if (timestamp, sensor_id) in merged_records_lidar:
                continue
            merged_records_lidar[(timestamp, sensor_id)] = filename
    return merged_records_lidar


def merge_records_wifi(records_wifi_list: List[Optional[kapture.RecordsWifi]]) -> kapture.RecordsWifi:
    """
    Merge several wifi records lists. For wifi record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_wifi_list: list of wifi records
    :return: merged wifi records
    """
    assert len(records_wifi_list) > 0

    merged_records_wifi = kapture.RecordsWifi()
    for records_wifi in records_wifi_list:
        if records_wifi is None:
            continue
        for timestamp, sensor_id, record_wifi in kapture.flatten(records_wifi):
            if (timestamp, sensor_id) in merged_records_wifi:
                continue
            merged_records_wifi[(timestamp, sensor_id)] = record_wifi
    return merged_records_wifi


def merge_records_gnss(records_gnss_list: List[Optional[kapture.RecordsGnss]]) -> kapture.RecordsGnss:
    """
    Merge several gnss records lists. For gnss record with the same timestamp and sensor identifier,
     keep only the first one.

    :param records_gnss_list: list of gnss records
    :return: merged gnss records
    """
    assert len(records_gnss_list) > 0

    merged_records_gnss = kapture.RecordsWifi()
    for records_gnss in records_gnss_list:
        if records_gnss is None:
            continue
        for timestamp, sensor_id, record_gnss in kapture.flatten(records_gnss):
            if (timestamp, sensor_id) in merged_records_gnss:
                continue
            merged_records_gnss[(timestamp, sensor_id)] = record_gnss
    return merged_records_gnss


def merge_keep_ids(kapture_list: List[kapture.Kapture], skip_list: List[Type],
                   data_paths: List[str], kapture_path: str,
                   images_import_method: TransferAction) -> kapture.Kapture:
    """
    Merge multiple kapture while keeping ids (sensor_id) identical in merged and inputs.

    :param kapture_list: list of kapture to merge.
    :param skip_list: optional types not to merge. sensors and rigs are unskipable
    :param data_paths: list of path to root path directory in same order as mentioned in kapture_list.
    :param kapture_path: directory root path to the merged kapture.
    :return: merged kapture
    """
    merged_kapture = kapture.Kapture()

    # get the union of all sensors
    new_sensors = merge_sensors([every_kapture.sensors for every_kapture in kapture_list])
    if new_sensors:  # if merge_sensors returned an empty object, keep merged_kapture.sensors to None
        merged_kapture.sensors = new_sensors

    # get the union of all rigs
    new_rigs = merge_rigs([every_kapture.rigs for every_kapture in kapture_list])
    if new_rigs:  # if merge_rigs returned an empty object, keep merged_kapture.sensors to None
        merged_kapture.rigs = new_rigs

    # all fields below can be skipped with skip_list
    # we do not assign the properties when the merge evaluate to false, we keep it as None
    if kapture.Trajectories not in skip_list:
        new_trajectories = merge_trajectories([every_kapture.trajectories for every_kapture in kapture_list])
        if new_trajectories:
            merged_kapture.trajectories = new_trajectories

    if kapture.RecordsCamera not in skip_list:
        new_records_camera = merge_records_camera([every_kapture.records_camera for every_kapture in kapture_list])
        if new_records_camera:
            merged_kapture.records_camera = new_records_camera

        merge_records_data([[image_name
                             for _, _, image_name in kapture.flatten(every_kapture.records_camera)]
                            if every_kapture.records_camera is not None else []
                            for every_kapture in kapture_list],
                           [get_image_fullpath(data_path, image_filename=None) for data_path in data_paths],
                           kapture_path,
                           images_import_method)

    if kapture.RecordsLidar not in skip_list:
        new_records_lidar = merge_records_lidar([every_kapture.records_lidar for every_kapture in kapture_list])
        if new_records_lidar:
            merged_kapture.records_lidar = new_records_lidar
    if kapture.RecordsWifi not in skip_list:
        new_records_wifi = merge_records_wifi([every_kapture.records_wifi for every_kapture in kapture_list])
        if new_records_wifi:
            merged_kapture.records_wifi = new_records_wifi
    if kapture.RecordsGnss not in skip_list:
        new_records_gnss = merge_records_gnss([every_kapture.records_gnss for every_kapture in kapture_list])
        if new_records_gnss:
            merged_kapture.records_gnss = new_records_gnss

    # for the reconstruction, except points and observations, the files are copied with shutil.copy
    # if kapture_path evaluates to False, all copies will be skipped (but classes will be filled normally)
    if kapture.Keypoints not in skip_list:
        keypoints = [every_kapture.keypoints for every_kapture in kapture_list]
        keypoints_not_none = [k for k in keypoints if k is not None]
        if len(keypoints_not_none) > 0:
            new_keypoints = merge_keypoints(keypoints, data_paths, kapture_path)
            if new_keypoints:
                merged_kapture.keypoints = new_keypoints
    if kapture.Descriptors not in skip_list:
        descriptors = [every_kapture.descriptors for every_kapture in kapture_list]
        descriptors_not_none = [k for k in descriptors if k is not None]
        if len(descriptors_not_none) > 0:
            new_descriptors = merge_descriptors(descriptors, data_paths, kapture_path)
            if new_descriptors:
                merged_kapture.descriptors = new_descriptors
    if kapture.GlobalFeatures not in skip_list:
        global_features = [every_kapture.global_features for every_kapture in kapture_list]
        global_features_not_none = [k for k in global_features if k is not None]
        if len(global_features_not_none) > 0:
            new_global_features = merge_global_features(global_features, data_paths, kapture_path)
            if new_global_features:
                merged_kapture.global_features = new_global_features
    if kapture.Matches not in skip_list:
        matches = [every_kapture.matches for every_kapture in kapture_list]
        new_matches = merge_matches(matches, data_paths, kapture_path)
        if new_matches:
            merged_kapture.matches = new_matches

    if kapture.Points3d not in skip_list and kapture.Observations not in skip_list:
        points_and_obs = [(every_kapture.points3d, every_kapture.observations) for every_kapture in kapture_list]
        new_points, new_observations = merge_points3d_and_observations(points_and_obs)
        if new_points:
            merged_kapture.points3d = new_points
        if new_observations:
            merged_kapture.observations = new_observations
    elif kapture.Points3d not in skip_list:
        points = [every_kapture.points3d for every_kapture in kapture_list]
        new_points = merge_points3d(points)
        if new_points:
            merged_kapture.points3d = new_points
    return merged_kapture



#!/usr/bin/env python3
# Copyright 2020-present NAVER Corp. Under BSD 3-clause license

"""
Kapture evaluation functions to save results on disk.
"""

import argparse
import logging
import os
import os.path as path
import numpy as np
from statistics import mean, median
from math import isnan
from typing import List, Tuple

import path_to_kapture
import kapture
import kapture.io.csv as csv
from kapture.algo.evaluation import evaluate, fill_bins
import kapture.utils.logging
from kapture.utils.paths import safe_remove_file

RESULTS_FILENAME = "scores"
STATISTICS_FILENAME = "stats"
PLOT_FILENAME = "plot"

logger = logging.getLogger('evaluate')


def write_results_to_file(output_folder: str,
                          labels: List[str],
                          results: List[List[Tuple[str, float, float]]],
                          force: bool) -> None:
    """
    Writes evaluation results to text files. Results and labels must be synchronised.

    :param output_folder: full path of folder to write files in.
    :param labels: labels for the result files
    :param results: results to write
    :param force: Silently overwrite results files if already exists.
    """
    for i in range(0, len(results)):
        label = labels[i]
        result = results[i]
        full_path = path.join(output_folder, RESULTS_FILENAME + '_{}'.format(label) + '.txt')
        safe_remove_file(full_path, force)
        results_as_lines = ['{} {} {}\n'.format(name, position_error, rotation_error)
                            for name, position_error, rotation_error in result]
        with open(full_path, 'w') as fid:
            fid.writelines(results_as_lines)


def write_statistics_to_file(output_folder: str,
                             labels: List[str],
                             results: List[List[Tuple[str, float, float]]],
                             bins_as_str: List[str],
                             force: bool) -> None:
    """
    Writes evaluation statistics to text files. Results and labels must be synchronised.

    :param output_folder: full path of folder to write statistics files in.
    :param labels: labels for the statistics files
    :param results: results to compute statistics
    :param bins_as_str: list of bin names
    :param force: Silently overwrite results files if already exists.
    """
    full_path = path.join(output_folder, STATISTICS_FILENAME + '.txt')
    safe_remove_file(full_path, force)

    bins = [(float(split_bin[0]), float(split_bin[1])) for split_bin in map(lambda x: x.split(), bins_as_str)]
    print_line = ''
    for i in range(0, len(results)):
        label = labels[i]
        result = results[i]
        number_of_images = len(result)

        positions_errors_all = [position_error if not isnan(position_error) else float("inf")
                                for name, position_error, rotation_error in result]
        rotation_errors_all = [rotation_error if not isnan(rotation_error) else float("inf")
                               for name, position_error, rotation_error in result]

        positions_errors = [position_error
                            for name, position_error, rotation_error in result if not isnan(position_error)]
        rotation_errors = [rotation_error
                           for name, position_error, rotation_error in result if not isnan(rotation_error)]

        print_line += 'Model: {}\n\n'.format(label)
        print_line += 'Found {} / {} image positions ({:.2f} %).\n'.format(
            len(positions_errors), number_of_images, float(100.0*len(positions_errors)/number_of_images))
        print_line += 'Found {} / {} image rotations ({:.2f} %).\n'.format(
            len(rotation_errors), number_of_images, float(100.0*len(rotation_errors)/number_of_images))

        print_line += 'Localized images: mean=({:.4f}m, {:.4f} deg) / median=({:.4f}m, {:.4f} deg)\n'.format(
            mean(positions_errors),
            mean(rotation_errors),
            median(positions_errors),
            median(rotation_errors))
        print_line += 'All: median=({:.4f}m, {:.4f} deg)\n'.format(median(positions_errors_all),
                                                                   median(rotation_errors_all))
        print_line += 'Min: {:.4f}m; {:.4f} deg\n'.format(min(positions_errors), min(rotation_errors))
        print_line += 'Max: {:.4f}m; {:.4f} deg\n\n'.format(max(positions_errors), max(rotation_errors))

        filled_bins = fill_bins(result, bins)
        bins_lines = ['({}m, {} deg): {:.2f}%\n'.format(
            position_error,
            rotation_error,
            (number_of_images_in_bin / number_of_images) * 100.0)
            for position_error, rotation_error, number_of_images_in_bin in filled_bins]
        print_line += ''.join(bins_lines)
        print_line += '\n'
    print(print_line)
    with open(full_path, 'w') as fid:
        fid.write(print_line)


def plot_localized_over_position_threshold(output_folder: str,
                                           labels: List[str],
                                           results: List[List[Tuple[str, float, float]]],
                                           rotation_threshold: float, plot_max: int,
                                           title: str,
                                           plot_loc: str, plot_font_size: int, plot_legend_font_size: int,
                                           force: bool) -> None:
    """
    Plot image with localized positions.

    :param output_folder: full path of folder to write plot files in.
    :param labels: labels for the plot images
    :param results: results to compute the images to plot
    :param rotation_threshold: rotation threshold
    :param plot_max: max points to plot
    :param title: title for the plotted image
    :param plot_loc: location for the legend
    :param plot_font_size: font size to plot with
    :param plot_legend_font_size: font size for the legend
    :param force: Silently overwrite plot files if already exists.
    """
    import matplotlib.pylab as plt
    plt.rcParams['font.size'] = plot_font_size
    plt.rcParams['legend.fontsize'] = plot_legend_font_size
    position_thresholds = np.linspace(0.0, 1.0, num=101, dtype=np.float64) * plot_max  # thresholds in cm
    bins = [(position_threshold / 100.0, rotation_threshold)
            for position_threshold in position_thresholds]  # convert to thresholds in m
    number_of_images = None
    for i in range(0, len(results)):
        label = labels[i]
        result = results[i]
        if number_of_images is None:
            number_of_images = len(result)
        else:
            # just make sure we are comparing comparable things
            assert(number_of_images == len(result))
        filled_bins = [(t[2] / number_of_images) * 100.0 for t in fill_bins(result, bins)]  # convert back to cm
        plt.plot(position_thresholds, filled_bins, lw=2, label=label)
    plt.xlabel('position error threshold (cm)')
    plt.ylabel('localized images (%)')
    plt.ylim([0, 100])  # 0 to 100 %
    plt.legend(loc=plot_loc)
    plt.xlim([position_thresholds[0], position_thresholds[-1]])
    # plot grid
    grid_step = int(plot_max / 10.0)
    for i in range(grid_step, plot_max - grid_step + 1, grid_step):
        plt.plot([i, i], [0, 100], 'k', alpha=0.1)
    for i in range(10, 91, 10):
        plt.plot([0, plot_max], [i, i], 'k', alpha=0.1)

    plot_title = title
    if rotation_threshold >= 0:
        plot_title += (' ' if title else '') + r'$\alpha_{th}$' + '={} deg'.format(rotation_threshold)
    plt.title(plot_title)

    full_path = path.join(output_folder, PLOT_FILENAME + '.png')
    safe_remove_file(full_path, force)
    plt.savefig(full_path, bbox_inches='tight')


def save_evaluation(results: List[List[Tuple[str, float, float]]],
                    output_dir: str,
                    labels: List[str],
                    bins_as_str: List[str],
                    rotation_threshold: float,
                    plot_max: int,
                    plot_title: str,
                    plot_loc: str, plot_font_size: int, plot_legend_font_size: int,
                    force: bool) -> None:
    """
    Save the evaluation results into results, statistics and plot files. Results and labels must be synchronised.

    :param results: results to save
    :param output_dir: full path of folder to write statistics files in.
    :param labels: labels of the results.
    :param bins_as_str: list of bin names
    :param rotation_threshold: rotation threshold for the plot
    :param plot_max: max points to plot
    :param plot_title: title for the plot
    :param plot_loc: location of the plot legend
    :param plot_font_size: font size to plot with
    :param plot_legend_font_size: font size for the legend
    :param force: Silently overwrite files if already exists.
    """
    write_results_to_file(output_dir, labels, results, force)
    write_statistics_to_file(output_dir, labels, results, bins_as_str, force)
    plot_localized_over_position_threshold(output_dir, labels, results,
                                           rotation_threshold, plot_max, plot_title,
                                           plot_loc, plot_font_size, plot_legend_font_size,
                                           force)


def evaluate_command_line() -> None:
    """
    Do the evaluation using the parameters given on the command line.
    """
    parser = argparse.ArgumentParser(description='Evaluation script for kapture data.')
    parser_verbosity = parser.add_mutually_exclusive_group()
    parser_verbosity.add_argument(
        '-v', '--verbose', nargs='?', default=logging.WARNING, const=logging.INFO,
        action=kapture.utils.logging.VerbosityParser,
        help='verbosity level (debug, info, warning, critical, ... or int value) [warning]')
    parser_verbosity.add_argument(
        '-q', '--silent', '--quiet', action='store_const', dest='verbose', const=logging.CRITICAL)
    parser.add_argument('-i', '--inputs', nargs='+',
                        help='input path to kapture data root directory. You can compare multiple models')
    parser.add_argument('--labels', nargs='+', default=[],
                        help='labels for inputs. must be of same length as inputs')
    parser.add_argument('-gt', '--ground-truth', required=True,
                        help='input path to data ground truth root directory in kapture format')
    parser.add_argument('-o', '--output',
                        help='output directory.', required=True)
    parser.add_argument('-l', '--image-list', default="",
                        help='optional, path to a text file containing the list of images to consider'
                        ' (1 line per image). if not present, all gt images are used')
    parser.add_argument('--bins', nargs='+', default=["0.25 2", "0.5 5", "5 10"],
                        help='the desired positions/rotations thresholds for bins'
                        'format is string : position_threshold_in_m space rotation_threshold_in_degree')
    parser.add_argument('-p', '--plot-rotation-threshold', default=-1, type=float,
                        help='rotation threshold for position error threshold plot. negative values -> ignore rotation')
    parser.add_argument('--plot-max', default=100, type=int, help='maximum distance in cm shown in plot')
    parser.add_argument('--plot-title', default="",
                        help='title for position error threshold plot')
    parser.add_argument('--plot-loc', default="best", choices=['best', 'upper right', 'upper left', 'lower left',
                                                               'lower right', 'right', 'center left', 'center right',
                                                               'lower center', 'upper center', 'center'],
                        help='position of plot legend. loc param for plt.legend.')
    parser.add_argument('--plot-font-size', default=15, type=int,
                        help='value for plt.rcParams[\'font.size\']')
    parser.add_argument('--plot-legend-font-size', default=8, type=int,
                        help='value for plt.rcParams[\'legend.fontsize\']')
    parser.add_argument('-f', '-y', '--force', action='store_true', default=False,
                        help='Force delete output directory if already exists')
    args = parser.parse_args()

    logger.setLevel(args.verbose)
    if args.verbose <= logging.DEBUG:
        # also let kapture express its logs
        kapture.utils.logging.getLogger().setLevel(args.verbose)
    assert(len(args.inputs) > 0)
    if len(args.labels) == 0:
        args.labels = [f'input{i}' for i in range(1, len(args.inputs) + 1)]
    assert(len(args.labels) == len(args.inputs))

    try:
        logger.debug(''.join(['\n\t{:13} = {}'.format(k, v)
                              for k, v in vars(args).items()]))

        os.makedirs(args.output, exist_ok=True)

        logger.debug('loading: {}'.format(args.inputs))
        all_kapture_to_eval = [csv.kapture_from_dir(folder) for folder in args.inputs]

        logger.info('loading ground truth data')
        gt_kapture = csv.kapture_from_dir(args.ground_truth)
        assert gt_kapture.records_camera is not None

        if args.image_list:
            with open(args.image_list, 'r') as fid:
                lines = fid.readlines()
                image_set = {line.rstrip() for line in lines}
        else:
            image_set = set(image_name for _, _, image_name in kapture.flatten(gt_kapture.records_camera))
        if len(image_set) == 0:
            logger.info('image_set is empty, for some reason, I could not find images to evaluate')
            exit(0)

        results = [evaluate(kapture_to_eval, gt_kapture, image_set) for kapture_to_eval in all_kapture_to_eval]

        save_evaluation(results, args.output, args.labels, args.bins,
                        args.plot_rotation_threshold, args.plot_max, args.plot_title,
                        args.plot_loc, args.plot_font_size, args.plot_legend_font_size,
                        args.force)

    except Exception as e:
        logger.critical(e)
        if args.verbose > 1:
            raise


if __name__ == '__main__':
    evaluate_command_line()
